/*
    Copyright (C) 2000 Red Hat, Inc.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/* By Elliot Lee <sopwith@redhat.com>, because some guy didn't bother
   to put a copyright/license on the previous rdate. See bugzilla.redhat.com,
   bug #8619. */
/* By Michael Tokarev <mjt@corpit.ru>: almost rewritten again,
   use average time from multiple hosts, print deltas, some safe
   controls, and some bugfixes compared to version by Elliot Lee */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdint.h>
#include <time.h>
#include <signal.h>
#include <stdarg.h>
#include <getopt.h>

/* difference between Unix time and net time */
#define BASE1970	2208988800UL
#define DEFAULT_PORT	37

static char *progname;

static void
print_err(const char *fmt, ...)
{
  va_list ap;
  fprintf(stderr, "%s: ", progname);
  va_start(ap,fmt);
  vfprintf(stderr,fmt,ap);
  va_end(ap);
  putc('\n', stderr);
}

/* Use TCP connect by default */
static int use_udp;
static short time_port;
static int timeout;

static volatile int alarmed;
static void
sigalarm(int sig)
{
  alarmed = 1;
}

struct htime {
  const char *ht_name;
  time_t ht_time;
  time_t ht_sent;
  time_t ht_recv;
};

static int
rdate(const char *hostname, struct htime *ht)
{
  struct sockaddr_in saddr;
  int fd;
  unsigned char time_buf[4];
  const char *action = NULL;
  int n;

  if (!inet_aton(hostname, &saddr.sin_addr)) {
    struct hostent *hent = gethostbyname(hostname);
    if(!hent) {
      print_err("unknown host %s: %s", hostname, hstrerror(h_errno));
      return -1;
    }
    if (hent->h_addrtype != AF_INET) {
      print_err("host %s: unsupported address type", hostname);
      return -1;
    }
    memcpy(&saddr.sin_addr, hent->h_addr, hent->h_length);
  }
  saddr.sin_family = AF_INET;
  saddr.sin_port = time_port;

  if (use_udp)
    fd = socket(AF_INET, SOCK_DGRAM, 0);
  else
    fd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);

  if (fd < 0) {
    print_err("couldn't create socket: %m");
    return -1;
  }

  if (timeout > 0) {
    alarmed = 0;
    alarm(timeout);
  }
  ht->ht_sent = time(NULL);
  if (use_udp) {
    if ((n = sendto(fd, NULL, 0, 0, &saddr, sizeof(saddr))) != 0)
      action = "sending to";
    else if ((n = recvfrom(fd, &time_buf, 4, 0, NULL, NULL)) != 4)
      action = "receiving from";
    else
      n = 0;
  }
  else {
    if ((n = connect(fd, &saddr, sizeof(saddr))) != 0)
      action = "connecting to";
    else {
      int r = 0;
      action = "reading from";
      do
        n = read(fd, time_buf + r, 4 - r);
      while(n > 0 && (r += n) < 4);
      if (n >= 0)
        n = !n;
    }
  }

  if (timeout > 0)
    alarm(0);
  if (!n) {
    ht->ht_recv = time(NULL);
    close(fd);
    /* See inetd's builtins.c for an explanation */
    ht->ht_time = (time_t)(ntohl(*(uint32_t *)time_buf) - BASE1970);
    ht->ht_name = hostname;
    return 0;
  }
  else {
    if (alarmed)
      print_err("timeout %s %s", action, hostname);
    else if (n < 0)
      print_err("error %s %s: %m", action, hostname);
    else
      print_err("error %s %s", action, hostname);
    close(fd);
    return -1; 
  }
}

static void
usage()
{
  print_err("try `%s --help' for help", progname);
  exit(1);
}

static int
natural(const char *str, const char *descr)
{
  char c;
  int n;
  if (sscanf(str, "%d%c", &n, &c) != 1) {
    print_err("invalid %s: `%s'", descr, str);
    exit(1);
  }
  return n;
}

int main(int argc, char *argv[])
{
  int i;
  int print = 0, printavg = 0, set = 0;
  int printdelta = 0, maxchange = 0, mincount = 1;
  int got = 0;
  struct htime *ht = NULL;
  time_t local, remote;
  int64_t sum;

  static const struct option opts[] = {
  { "help", 0, 0, 'h' },
  { "udp", 0, 0, 'u' },
  { "timeout", 1, 0, 't' },
  { "print", 0, 0, 'p' },
  { "avg", 0, 0, 'a' },
  { "average", 0, 0, 'a' },
  { "delta", 0, 0, 'd' },
  { "set", 0, 0, 's' },
  { "maxchange", 1, 0, 'm' },
  { "mincount", 1, 0, 'c' },
  { NULL, 0, 0, 0 }
  };

  if ((progname = strrchr(argv[0], '/')) != NULL)
    argv[0] = ++progname;
  else
    progname = argv[0];

  while((i = getopt_long(argc, argv, "ut:padsm:c:h", opts, NULL)) != EOF)
    switch(i) {
    case 'u': use_udp = 1; break;
    case 't': timeout = natural(optarg, "timeout"); break;
    case 'p': print = 1; break;
    case 'a': printavg = 1; break;
    case 'd': printdelta = 1; break;
    case 's': set = 1; break;
    case 'm': maxchange = natural(optarg, "maxchange"); break;
    case 'c': mincount = natural(optarg, "mincount"); break;
    case 'h':
      printf("\
%s: print or set time based on remote time\n\
Usage: %s [options] host..., where options are:\n\
 --help, -h -- print this help and exit\n\
 --udp, -u -- use UDP instead of TCP\n\
 --timeout, -t sec -- timeout for one host (0)\n\
 --print, -p -- print remote time (default)\n\
 --average, --avg, -a -- print the average remote time\n\
  --delta, -d -- print delta between local time
 --set, -s -- set system time based on remote time\n\
  --maxchange, -m sec -- max local time change allowed (0)\n\
  --mincount, -c count -- min answers to set time (1)\n\
", progname, progname);
      return 0;
    default:
      usage();
    }

  argv += optind; argc -= optind;

  if (!argc)
    usage();

  if (!set && !print && !printavg)
    print = 1;

  { /* find port to use */
    const char *proto = use_udp ? "udp" : "tcp";
    struct servent *sent = getservbyname("time", proto);
    if (sent)
      time_port = sent->s_port;
    else {
      /* fprintf(stderr,
              "%s: warning: time/%s: service not found, using default %d\n",
              progname, proto, DEFAULT_PORT); */
      time_port = htons(DEFAULT_PORT);
    }
  }

  if ((ht = (struct htime*)calloc((argc + 1), sizeof(struct htime))) == NULL) {
    print_err("unable to allocate memory");
    return 1;
  }

  if (timeout > 0) {
    struct sigaction sa;
    sa.sa_handler = sigalarm;
    sigemptyset(&sa.sa_mask);
    sa.sa_flags = 0;
    sigaction(SIGALRM, &sa, NULL);
  }

  for (i = 0; i < argc; ++i)
    if (rdate(argv[i], &ht[got]) == 0)
      ++got;

  if (!got)
    return 1;

  /* normalize time */
  for (local = time(NULL), sum = 0, i = 0; i < got; ++i) {
    ht[i].ht_time +=
      (local - ht[i].ht_sent) + (ht[i].ht_recv - ht[i].ht_sent) / 2;
    sum += ht[i].ht_time;
  }
  remote = (time_t)(sum / got);

  if (set) {
    if (got < mincount)
      print_err("got only %d answers: time not set", got);
    else if (maxchange &&
             (local > remote ? local - remote : remote - local) > maxchange)
      print_err("local time change (%+d sec) is too big: time not set",
                remote - local);
    else if (stime(&remote) != 0)
      print_err("unable to set time: %m");
    else
      set = 0;
  }
  if (print)
    for (i = 0; i < got; ++i) {
      printf("[%s]\t%.24s", ht[i].ht_name, ctime(&ht[i].ht_time));
      if (printdelta)
	printf(" (%d sec)\n", (int)(ht[i].ht_time - local));
      else
	putchar('\n');
    }
  if (printavg) {
    printf("average\t%.24s", ctime(&remote));
    if (printdelta)
      printf(" (%d sec)\n", (int)(remote - local));
    else
      putchar('\n');
  }

  return set;
}

